package solution;

import java.util.*;
import problem.*;
//import solution.*;

public class RoboMover {
    
    private List<ArmConfig> universe1;
    private List<ArmConfig> universe2;
	
    /* A PRNG for determining the possible armconfigs we're moving to */
    private ConfigGenerator generator;
    /* Our objects we need for collision testing */
	private List<Obstacle> objects;
	/* Our tester is our collision detector */
    private ArmConfig goal;
    private ProblemSpec ps;
    private LocalPlanner lp;
    
    /**
     * Constructor for RoboMover
     * 
     * @param ps is the problem spec that we are getting our obstacles from.
     */
    public RoboMover(ProblemSpec ps) {
    	objects = ps.getObstacles();
    	generator = new ConfigGenerator(objects);
    	goal = ps.getGoalState();
    	this.ps = ps;
    	lp = new LocalPlanner();
    	lp.setObjects(ps.getObstacles());
    	universe1 = generator.generateObstacleUniverse(ps);  	
    	universe2 = generator.generateRandomUniverse(ps);  	
    	System.out.println(String.format("Universe generated... %d points", universe1.size() + universe2.size()));
    }

	/** 
     * ALL the complexity of going from one node to the next is contained in this method
     * It will construct a valid path going from initialNode to nextNode or it will return
     * null if no path exists according to this class. If you want to move to a node 
     * just call this function.
     * 
     * @param An initial node
     * @param a goal node
     * 
     * @return a path from the current node to the next one. The current node will be at
     * the bottom of the stack. The robot's current position (goalNode) will be at path.peek()
     */
    public Stack<ArmConfig> moveToNode(ArmConfig initialNode, ArmConfig goal) {
    	Stack<ArmConfig> path = new Stack<>();
    	List<ArmConfig> possibilities = new ArrayList<>();
 
    	//first we try the three ways of moving naively to the smart goals
    	path.push(initialNode);
    	if (lp.getThere(path, goal)) {
    		return path;
    	}
    	
    	//Smart Goals did not work so now we try different goal configs. 
    	//Same base location but different arm twists
    	possibilities = generator.getPossibleConfigs(goal);
    	for (ArmConfig possibleGoal : possibilities) {
    	    if (lp.getThere(path, possibleGoal)) {
    	    	lp.getThere(path, goal);
    	        return path;
    	    }
    	}
    	
    	//if that didn't work we try something trickier
    	if (AStar(path, goal)) {
    		lp.getThere(path, goal);
    		return path;
    	}
    	
    	return path;
    }
    
	

    
    private boolean AStar(Stack<ArmConfig> path, ArmConfig newGoal) {
    	List<JoeyNode> nodeList = new ArrayList<>();
    	List<JoeyNode> solvedTree = new ArrayList<>();
    	JoeyNode initial = new JoeyNode(path.peek(), newGoal, ps);
    	JoeyNode goal = new JoeyNode(newGoal, newGoal, ps);
    	JoeyNode dest;
    	
    	//set up two points in our graph we care about
    	nodeList.add(initial);
    	nodeList.add(goal);
    	//set up initial and goal as enemies
    	initial.getTo(goal);
    	while (true) {
	    	//The first thing we need is a bunch of nodes to try and navigate through
	    	sampleNodes(nodeList);
	    	nodeList.get(0).lookForFriends(nodeList);
	    	nodeList.get(1).lookForFriends(nodeList);
	    	//We need to find out who has friends and enemies
	    	System.out.println("Building Graph");
	    	generateGraph(nodeList);
	    	System.out.println("Searching Graph");
	    	//search graph
	    	dest = searchGraph(nodeList);
	
	    	if (dest == null) {
	    		System.out.println("I failed :'(... let's try again");
	    		continue;
	    	} 
    		while (dest != null) {
    			solvedTree.add(dest);
    			dest = dest.getParents();
    		}
    		
    		Collections.reverse(solvedTree);
    		for (JoeyNode level : solvedTree) {
    			lp.getThere(path, level.getNode());
    		}
    		return true;
    	}
	}

	private JoeyNode searchGraph(List<JoeyNode> nodeList) {
		PriorityQueue<JoeyNode> tree = new PriorityQueue<>();
		Set<JoeyNode> visited = new HashSet<>();
		JoeyNode current;
		
		tree.add(nodeList.get(0));
		while (tree.peek() != null) {
			Set<JoeyNode> expansion;
			Set<JoeyNode> expansion2;
			
			//We get the cheapest node in queue
			current = tree.poll();
			//If we've already been here ignore it
			if (visited.contains(current)) {
				continue;
			}
			//If not we say we have for next time
			visited.add(current);
			//check if we are done
			if (current.leave().contains(nodeList.get(1))) {
				System.out.println("Found it :)");
				return current;
			}
			//If we are not done, then we must expand our node.
			expansion = current.leave();
			expansion2 = new HashSet<>();
			expansion2.addAll(expansion);
			expansion.remove(current);
			
			for (JoeyNode peer : expansion) {
				if (visited.contains(peer)) {
					expansion2.remove(peer);
				}
			}
			
			for (JoeyNode member : expansion2) {
				member.arrive(current, current.edgeSize(member));
			}
			tree.addAll(expansion);
		}
		current = tree.poll();
		return current;
	}

	private void generateGraph(List<JoeyNode> nodeList) {
		for (JoeyNode peer : nodeList) {
			if (peer.popularity() < 5) {
				peer.lookForFriends(nodeList);
			}
		}
	}

	private void sampleNodes(List<JoeyNode> nodeList) {
		for (int i = 0; i < 50; i++) {
			JoeyNode newFriend;
			int index = (int) Math.floor(Math.random() * universe1.size());	
			newFriend = new JoeyNode(universe1.get(index), goal, ps);
			nodeList.add(newFriend);
		}
		for (int i = 0; i < 50; i++) {
			JoeyNode newFriend;
			int index = (int) Math.floor(Math.random() * universe2.size());	
			newFriend = new JoeyNode(universe2.get(index), goal, ps);
			nodeList.add(newFriend);
		}
	}
}	

